# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1alpha1ServingRuntimePodSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'affinity': 'V1Affinity',
        'containers': 'list[V1alpha1Container]',
        'node_selector': 'dict(str, str)',
        'tolerations': 'list[V1Toleration]'
    }

    attribute_map = {
        'affinity': 'affinity',
        'containers': 'containers',
        'node_selector': 'nodeSelector',
        'tolerations': 'tolerations'
    }

    def __init__(self, affinity=None, containers=None, node_selector=None, tolerations=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1ServingRuntimePodSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._affinity = None
        self._containers = None
        self._node_selector = None
        self._tolerations = None
        self.discriminator = None

        if affinity is not None:
            self.affinity = affinity
        self.containers = containers
        if node_selector is not None:
            self.node_selector = node_selector
        if tolerations is not None:
            self.tolerations = tolerations

    @property
    def affinity(self):
        """Gets the affinity of this V1alpha1ServingRuntimePodSpec.  # noqa: E501


        :return: The affinity of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :rtype: V1Affinity
        """
        return self._affinity

    @affinity.setter
    def affinity(self, affinity):
        """Sets the affinity of this V1alpha1ServingRuntimePodSpec.


        :param affinity: The affinity of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :type: V1Affinity
        """

        self._affinity = affinity

    @property
    def containers(self):
        """Gets the containers of this V1alpha1ServingRuntimePodSpec.  # noqa: E501

        List of containers belonging to the pod. Containers cannot currently be added or removed. There must be at least one container in a Pod. Cannot be updated.  # noqa: E501

        :return: The containers of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :rtype: list[V1alpha1Container]
        """
        return self._containers

    @containers.setter
    def containers(self, containers):
        """Sets the containers of this V1alpha1ServingRuntimePodSpec.

        List of containers belonging to the pod. Containers cannot currently be added or removed. There must be at least one container in a Pod. Cannot be updated.  # noqa: E501

        :param containers: The containers of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :type: list[V1alpha1Container]
        """
        if self.local_vars_configuration.client_side_validation and containers is None:  # noqa: E501
            raise ValueError("Invalid value for `containers`, must not be `None`")  # noqa: E501

        self._containers = containers

    @property
    def node_selector(self):
        """Gets the node_selector of this V1alpha1ServingRuntimePodSpec.  # noqa: E501

        NodeSelector is a selector which must be true for the pod to fit on a node. Selector which must match a node's labels for the pod to be scheduled on that node. More info: https://kubernetes.io/docs/concepts/configuration/assign-pod-node/  # noqa: E501

        :return: The node_selector of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._node_selector

    @node_selector.setter
    def node_selector(self, node_selector):
        """Sets the node_selector of this V1alpha1ServingRuntimePodSpec.

        NodeSelector is a selector which must be true for the pod to fit on a node. Selector which must match a node's labels for the pod to be scheduled on that node. More info: https://kubernetes.io/docs/concepts/configuration/assign-pod-node/  # noqa: E501

        :param node_selector: The node_selector of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :type: dict(str, str)
        """

        self._node_selector = node_selector

    @property
    def tolerations(self):
        """Gets the tolerations of this V1alpha1ServingRuntimePodSpec.  # noqa: E501

        If specified, the pod's tolerations.  # noqa: E501

        :return: The tolerations of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :rtype: list[V1Toleration]
        """
        return self._tolerations

    @tolerations.setter
    def tolerations(self, tolerations):
        """Sets the tolerations of this V1alpha1ServingRuntimePodSpec.

        If specified, the pod's tolerations.  # noqa: E501

        :param tolerations: The tolerations of this V1alpha1ServingRuntimePodSpec.  # noqa: E501
        :type: list[V1Toleration]
        """

        self._tolerations = tolerations

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1ServingRuntimePodSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1ServingRuntimePodSpec):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import re
from unittest import mock

import avro.io
import avro.schema
import io
import pytest
from cloudevents.http import CloudEvent, to_binary, to_structured
from kserve import Model
from kserve import ModelServer
from kserve import ModelRepository
from tornado.httpclient import HTTPClientError
from ray import serve


test_avsc_schema = '''
        {
        "namespace": "example.avro",
         "type": "record",
         "name": "User",
         "fields": [
             {"name": "name", "type": "string"},
             {"name": "favorite_number",  "type": ["int", "null"]},
             {"name": "favorite_color", "type": ["string", "null"]}
         ]
        }
        '''


def dummy_cloud_event(data, set_contenttype=False, add_extension=False):
    # This data defines a binary cloudevent
    attributes = {
        "type": "com.example.sampletype1",
        "source": "https://example.com/event-producer",
        "specversion": "1.0",
        "id": "36077800-0c23-4f38-a0b4-01f4369f670a",
        "time": "2021-01-28T21:04:43.144141+00:00"
    }
    if set_contenttype:
        attributes["content-type"] = "application/json"
    if add_extension:
        attributes["custom-extension"] = "custom-value"

    event = CloudEvent(attributes, data)
    return event


class DummyModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request):
        return {"predictions": request["instances"]}

    async def explain(self, request):
        return {"predictions": request["instances"]}


@serve.deployment
class DummyServeModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request):
        return {"predictions": request["instances"]}

    async def explain(self, request):
        return {"predictions": request["instances"]}


class DummyCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request):
        return {"predictions": request["instances"]}

    async def explain(self, request):
        return {"predictions": request["instances"]}


class DummyAvroCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    def _parserequest(self, request):
        schema = avro.schema.parse(test_avsc_schema)
        raw_bytes = request
        bytes_reader = io.BytesIO(raw_bytes)
        decoder = avro.io.BinaryDecoder(bytes_reader)
        reader = avro.io.DatumReader(schema)
        record1 = reader.read(decoder)
        return record1

    def preprocess(self, request):
        if isinstance(request, CloudEvent):
            attributes = request._attributes
            assert attributes["specversion"] == "1.0"
            assert attributes["source"] == "https://example.com/event-producer"
            assert attributes["type"] == "com.example.sampletype1"
            assert attributes["datacontenttype"] == "application/x-www-form-urlencoded"
            assert attributes["content-type"] == "application/json"
            return self._parserequest(request.data)

    async def predict(self, request):
        return {"predictions": [[request['name'], request['favorite_number'], request['favorite_color']]]}

    async def explain(self, request):
        return {"predictions": [[request['name'], request['favorite_number'], request['favorite_color']]]}


class DummyModelRepository(ModelRepository):
    def __init__(self, test_load_success: bool):
        super().__init__()
        self.test_load_success = test_load_success

    async def load(self, name: str) -> bool:
        if self.test_load_success:
            model = DummyModel(name)
            model.load()
            self.update(model)
            return model.ready
        else:
            return False


class TestTFHttpServer:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        model.load()
        server = ModelServer()
        server.register_model(model)
        return server.create_application()

    async def test_liveness(self, http_server_client):
        resp = await http_server_client.fetch('/')
        assert resp.code == 200
        assert resp.body == b'{"status": "alive"}'

    async def test_model(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel')
        assert resp.code == 200

    async def test_unknown_model(self, http_server_client):
        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/v1/models/InvalidModel')
        assert err.value.code == 404
        assert err.value.response.body == b'{"error": "Model with name InvalidModel does not exist."}'

    async def test_predict(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              body=b'{"instances":[[1,2]]}')
        assert resp.code == 200
        assert resp.body == b'{"predictions": [[1, 2]]}'
        assert resp.headers['content-type'] == "application/json; charset=UTF-8"

    async def test_explain(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel:explain',
                                              method="POST",
                                              body=b'{"instances":[[1,2]]}')
        assert resp.code == 200
        assert resp.body == b'{"predictions": [[1, 2]]}'
        assert resp.headers['content-type'] == "application/json; charset=UTF-8"

    async def test_list(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models')
        assert resp.code == 200
        assert resp.body == b'{"models": ["TestModel"]}'

    async def test_unknown_path(self, http_server_client):
        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/unknown_path')
        assert err.value.code == 404
        assert err.value.response.body == b'{"error": "invalid path"}'


class TestRayServer:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        serve.start(detached=False, http_options={"host": "0.0.0.0", "port": 9071})

        DummyServeModel.deploy("TestModel")
        handle = DummyServeModel.get_handle()
        handle.load.remote()

        server = ModelServer()
        server.register_model_handle("TestModel", handle)

        return server.create_application()

    async def test_liveness_handler(self, http_server_client):
        resp = await http_server_client.fetch('/')
        assert resp.code == 200
        assert resp.body == b'{"status": "alive"}'

    async def test_list_handler(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models')
        assert resp.code == 200
        assert resp.body == b'{"models": ["TestModel"]}'

    async def test_health_handler(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel')
        assert resp.code == 200
        assert resp.body == b'{"name": "TestModel", "ready": true}'

    async def test_predict(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              body=b'{"instances":[[1,2]]}')
        assert resp.code == 200
        assert resp.body == b'{"predictions": [[1, 2]]}'
        assert resp.headers['content-type'] == "application/json; charset=UTF-8"

    async def test_explain(self, http_server_client):
        resp = await http_server_client.fetch('/v1/models/TestModel:explain',
                                              method="POST",
                                              body=b'{"instances":[[1,2]]}')
        assert resp.code == 200
        assert resp.body == b'{"predictions": [[1, 2]]}'
        assert resp.headers['content-type'] == "application/json; charset=UTF-8"


class TestTFHttpServerLoadAndUnLoad:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        server = ModelServer(registered_models=DummyModelRepository(test_load_success=True))
        return server.create_application()

    async def test_load(self, http_server_client):
        resp = await http_server_client.fetch('/v2/repository/models/model/load',
                                              method="POST", body=b'')
        assert resp.code == 200
        assert resp.body == b'{"name": "model", "load": true}'

    async def test_unload(self, http_server_client):
        resp = await http_server_client.fetch('/v2/repository/models/model/unload',
                                              method="POST", body=b'')
        assert resp.code == 200
        assert resp.body == b'{"name": "model", "unload": true}'


class TestTFHttpServerLoadAndUnLoadFailure:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        server = ModelServer(registered_models=DummyModelRepository(test_load_success=False))
        return server.create_application()

    async def test_load_fail(self, http_server_client):
        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/v2/repository/models/model/load',
                                               method="POST", body=b'')
        assert err.value.code == 503

    async def test_unload_fail(self, http_server_client):
        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/v2/repository/models/model/unload',
                                               method="POST", body=b'')
        assert err.value.code == 404


class TestTFHttpServerModelNotLoaded:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        return server.create_application()

    async def test_model_not_ready_error(self, http_server_client):
        with pytest.raises(HTTPClientError) as excinfo:
            _ = await http_server_client.fetch('/v1/models/TestModel')
        assert excinfo.value.code == 503


class TestTFHttpServerCloudEvent:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyCEModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        return server.create_application()

    async def test_predict_ce_structured(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]})
        headers, body = to_structured(event)

        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              headers=headers,
                                              body=body)
        body = json.loads(resp.body)

        assert resp.code == 200
        assert resp.headers['content-type'] == "application/cloudevents+json"

        assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert body["data"] == {"predictions": [[1, 2]]}
        assert body['specversion'] == "1.0"
        assert body['source'] == "io.kserve.kfserver.TestModel"
        assert body['type'] == "io.kserve.inference.response"
        assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    async def test_predict_custom_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ,
                             {"CE_SOURCE": "io.kserve.kfserver.CustomSource", "CE_TYPE": "io.kserve.custom_type"}):
            event = dummy_cloud_event({"instances": [[1, 2]]})
            headers, body = to_structured(event)

            resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                                  method="POST",
                                                  headers=headers,
                                                  body=body)
            body = json.loads(resp.body)

            assert resp.code == 200
            assert resp.headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.kfserver.CustomSource"
            assert body['type'] == "io.kserve.custom_type"

    async def test_predict_merge_structured_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, add_extension=True)
            headers, body = to_structured(event)

            resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                                  method="POST",
                                                  headers=headers,
                                                  body=body)
            body = json.loads(resp.body)

            assert resp.code == 200
            assert resp.headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.kfserver.TestModel"
            assert body['type'] == "io.kserve.inference.response"
            assert body["custom-extension"] == "custom-value"  # Added by add_extension=True in dummy_cloud_event
            assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    async def test_predict_merge_binary_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True, add_extension=True)
            headers, body = to_binary(event)

            resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                                  method="POST",
                                                  headers=headers,
                                                  body=body)

            assert resp.code == 200
            assert resp.headers['content-type'] == "application/json"
            assert resp.headers['ce-specversion'] == "1.0"
            assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            # Added by add_extension=True in dummy_cloud_event
            assert resp.headers['ce-custom-extension'] == 'custom-value'
            assert resp.headers['ce-source'] == "io.kserve.kfserver.TestModel"
            assert resp.headers['ce-type'] == "io.kserve.inference.response"
            assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
            assert resp.body == b'{"predictions": [[1, 2]]}'

    async def test_predict_ce_binary_dict(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True)
        headers, body = to_binary(event)
        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              headers=headers,
                                              body=body)

        assert resp.code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.kfserver.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.body == b'{"predictions": [[1, 2]]}'

    async def test_predict_ce_binary_bytes(self, http_server_client):
        event = dummy_cloud_event(b'{"instances":[[1,2]]}', set_contenttype=True)
        headers, body = to_binary(event)
        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              headers=headers,
                                              body=body)

        assert resp.code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.kfserver.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.body == b'{"predictions": [[1, 2]]}'

    async def test_predict_ce_bytes_bad_format_exception(self, http_server_client):
        event = dummy_cloud_event(b'{', set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/v1/models/TestModel:predict',
                                               method="POST",
                                               headers=headers,
                                               body=body)
        assert err.value.code == 400
        error_regex = re.compile("Failed to decode or parse binary json cloudevent: "
                                 "Expecting property name enclosed in double quotes.*")
        response = json.loads(err.value.response.body)
        assert error_regex.match(response["error"]) is not None

    async def test_predict_ce_bytes_bad_hex_format_exception(self, http_server_client):
        event = dummy_cloud_event(b'0\x80\x80\x06World!\x00\x00', set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(HTTPClientError) as err:
            _ = await http_server_client.fetch('/v1/models/TestModel:predict',
                                               method="POST",
                                               headers=headers,
                                               body=body)
        assert err.value.code == 400
        error_regex = re.compile("Failed to decode or parse binary json cloudevent: "
                                 "'utf-8' codec can't decode byte 0x80 in position 1: invalid start byte.*")
        response = json.loads(err.value.response.body)
        assert error_regex.match(response["error"]) is not None


class TestTFHttpServerAvroCloudEvent:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyAvroCEModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        return server.create_application()

    async def test_predict_ce_avro_binary(self, http_server_client):
        schema = avro.schema.parse(test_avsc_schema)
        msg = {"name": "foo", "favorite_number": 1, "favorite_color": "pink"}

        writer = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer.write(msg, encoder)
        data = bytes_writer.getvalue()

        event = dummy_cloud_event(data, set_contenttype=True)
        # Creates the HTTP request representation of the CloudEvent in binary content mode
        headers, body = to_binary(event)
        resp = await http_server_client.fetch('/v1/models/TestModel:predict',
                                              method="POST",
                                              headers=headers,
                                              body=body)

        assert resp.code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.kfserver.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.body == b'{"predictions": [["foo", 1, "pink"]]}'



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1alpha1_serving_runtime_list import V1alpha1ServingRuntimeList  # noqa: E501
from kserve.rest import ApiException

class TestV1alpha1ServingRuntimeList(unittest.TestCase):
    """V1alpha1ServingRuntimeList unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1alpha1ServingRuntimeList
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # model = kserve.models.v1alpha1_serving_runtime_list.V1alpha1ServingRuntimeList()  # noqa: E501
        if include_optional :
            return V1alpha1ServingRuntimeList(
                api_version = '0', 
                items = [
                    kserve.models.v1alpha1_serving_runtime.V1alpha1ServingRuntime(
                        api_version = '0', 
                        kind = '0', 
                        metadata = None, 
                        spec = kserve.models.v1alpha1_serving_runtime_spec.V1alpha1ServingRuntimeSpec(
                            affinity = None, 
                            built_in_adapter = kserve.models.v1alpha1_built_in_adapter.V1alpha1BuiltInAdapter(
                                mem_buffer_bytes = 56, 
                                model_loading_timeout_millis = 56, 
                                runtime_management_port = 56, 
                                server_type = '0', ), 
                            containers = [
                                kserve.models.v1alpha1_container.V1alpha1Container(
                                    args = [
                                        '0'
                                        ], 
                                    command = [
                                        '0'
                                        ], 
                                    env = [
                                        None
                                        ], 
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    working_dir = '0', )
                                ], 
                            disabled = True, 
                            grpc_data_endpoint = '0', 
                            grpc_endpoint = '0', 
                            http_data_endpoint = '0', 
                            node_selector = {
                                'key' : '0'
                                }, 
                            replicas = 56, 
                            storage_helper = kserve.models.v1alpha1_storage_helper.V1alpha1StorageHelper(
                                disabled = True, ), 
                            supported_model_formats = [
                                kserve.models.v1alpha1_supported_model_format.V1alpha1SupportedModelFormat(
                                    auto_select = True, 
                                    name = '0', 
                                    version = '0', )
                                ], 
                            tolerations = [
                                None
                                ], ), 
                        status = None, )
                    ], 
                kind = '0', 
                metadata = None
            )
        else :
            return V1alpha1ServingRuntimeList(
                items = [
                    kserve.models.v1alpha1_serving_runtime.V1alpha1ServingRuntime(
                        api_version = '0', 
                        kind = '0', 
                        metadata = None, 
                        spec = kserve.models.v1alpha1_serving_runtime_spec.V1alpha1ServingRuntimeSpec(
                            affinity = None, 
                            built_in_adapter = kserve.models.v1alpha1_built_in_adapter.V1alpha1BuiltInAdapter(
                                mem_buffer_bytes = 56, 
                                model_loading_timeout_millis = 56, 
                                runtime_management_port = 56, 
                                server_type = '0', ), 
                            containers = [
                                kserve.models.v1alpha1_container.V1alpha1Container(
                                    args = [
                                        '0'
                                        ], 
                                    command = [
                                        '0'
                                        ], 
                                    env = [
                                        None
                                        ], 
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    working_dir = '0', )
                                ], 
                            disabled = True, 
                            grpc_data_endpoint = '0', 
                            grpc_endpoint = '0', 
                            http_data_endpoint = '0', 
                            node_selector = {
                                'key' : '0'
                                }, 
                            replicas = 56, 
                            storage_helper = kserve.models.v1alpha1_storage_helper.V1alpha1StorageHelper(
                                disabled = True, ), 
                            supported_model_formats = [
                                kserve.models.v1alpha1_supported_model_format.V1alpha1SupportedModelFormat(
                                    auto_select = True, 
                                    name = '0', 
                                    version = '0', )
                                ], 
                            tolerations = [
                                None
                                ], ), 
                        status = None, )
                    ],
        )

    def testV1alpha1ServingRuntimeList(self):
        """Test V1alpha1ServingRuntimeList"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import tornado.web
from http import HTTPStatus
from kserve.model_repository import ModelRepository


class BaseHandler(tornado.web.RequestHandler):
    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """This method is called when there are unhandled tornado.web.HTTPErrors"""
        self.set_status(status_code)

        reason = "An error occurred"

        exc_info = kwargs.get("exc_info", None)
        if exc_info is not None:
            if hasattr(exc_info[1], "reason"):
                reason = exc_info[1].reason
            else:
                reason = str(exc_info[1])

        self.write({"error": reason})


class NotFoundHandler(tornado.web.RequestHandler):
    def write_error(self, status_code: int, **kwargs: Any) -> None:
        self.set_status(HTTPStatus.NOT_FOUND)
        self.write({"error": "invalid path"})


class HTTPHandler(BaseHandler):
    def initialize(self, models: ModelRepository):
        self.models = models  # pylint:disable=attribute-defined-outside-init

    def get_model(self, name: str):
        model = self.models.get_model(name)
        if model is None:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.NOT_FOUND,
                reason="Model with name %s does not exist." % name
            )
        if not self.models.is_model_ready(name):
            model.load()
        return model



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

# flake8: noqa
"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from kserve.models.v1alpha1_built_in_adapter import V1alpha1BuiltInAdapter
from kserve.models.v1alpha1_cluster_serving_runtime import V1alpha1ClusterServingRuntime
from kserve.models.v1alpha1_cluster_serving_runtime_list import V1alpha1ClusterServingRuntimeList
from kserve.models.v1alpha1_container import V1alpha1Container
from kserve.models.v1alpha1_model_spec import V1alpha1ModelSpec
from kserve.models.v1alpha1_serving_runtime import V1alpha1ServingRuntime
from kserve.models.v1alpha1_serving_runtime_list import V1alpha1ServingRuntimeList
from kserve.models.v1alpha1_serving_runtime_pod_spec import V1alpha1ServingRuntimePodSpec
from kserve.models.v1alpha1_serving_runtime_spec import V1alpha1ServingRuntimeSpec
from kserve.models.v1alpha1_storage_helper import V1alpha1StorageHelper
from kserve.models.v1alpha1_supported_model_format import V1alpha1SupportedModelFormat
from kserve.models.v1alpha1_trained_model import V1alpha1TrainedModel
from kserve.models.v1alpha1_trained_model_list import V1alpha1TrainedModelList
from kserve.models.v1alpha1_trained_model_spec import V1alpha1TrainedModelSpec
from kserve.models.v1beta1_aix_explainer_spec import V1beta1AIXExplainerSpec
from kserve.models.v1beta1_art_explainer_spec import V1beta1ARTExplainerSpec
from kserve.models.v1beta1_alibi_explainer_spec import V1beta1AlibiExplainerSpec
from kserve.models.v1beta1_batcher import V1beta1Batcher
from kserve.models.v1beta1_component_extension_spec import V1beta1ComponentExtensionSpec
from kserve.models.v1beta1_component_status_spec import V1beta1ComponentStatusSpec
from kserve.models.v1beta1_custom_explainer import V1beta1CustomExplainer
from kserve.models.v1beta1_custom_predictor import V1beta1CustomPredictor
from kserve.models.v1beta1_custom_transformer import V1beta1CustomTransformer
from kserve.models.v1beta1_deploy_config import V1beta1DeployConfig
from kserve.models.v1beta1_explainer_config import V1beta1ExplainerConfig
from kserve.models.v1beta1_explainer_extension_spec import V1beta1ExplainerExtensionSpec
from kserve.models.v1beta1_explainer_spec import V1beta1ExplainerSpec
from kserve.models.v1beta1_explainers_config import V1beta1ExplainersConfig
from kserve.models.v1beta1_inference_service import V1beta1InferenceService
from kserve.models.v1beta1_inference_service_list import V1beta1InferenceServiceList
from kserve.models.v1beta1_inference_service_spec import V1beta1InferenceServiceSpec
from kserve.models.v1beta1_inference_service_status import V1beta1InferenceServiceStatus
from kserve.models.v1beta1_inference_services_config import V1beta1InferenceServicesConfig
from kserve.models.v1beta1_ingress_config import V1beta1IngressConfig
from kserve.models.v1beta1_light_gbm_spec import V1beta1LightGBMSpec
from kserve.models.v1beta1_logger_spec import V1beta1LoggerSpec
from kserve.models.v1beta1_model_format import V1beta1ModelFormat
from kserve.models.v1beta1_model_spec import V1beta1ModelSpec
from kserve.models.v1beta1_onnx_runtime_spec import V1beta1ONNXRuntimeSpec
from kserve.models.v1beta1_pmml_spec import V1beta1PMMLSpec
from kserve.models.v1beta1_paddle_server_spec import V1beta1PaddleServerSpec
from kserve.models.v1beta1_pod_spec import V1beta1PodSpec
from kserve.models.v1beta1_predictor_config import V1beta1PredictorConfig
from kserve.models.v1beta1_predictor_extension_spec import V1beta1PredictorExtensionSpec
from kserve.models.v1beta1_predictor_protocols import V1beta1PredictorProtocols
from kserve.models.v1beta1_predictor_spec import V1beta1PredictorSpec
from kserve.models.v1beta1_predictors_config import V1beta1PredictorsConfig
from kserve.models.v1beta1_sk_learn_spec import V1beta1SKLearnSpec
from kserve.models.v1beta1_tf_serving_spec import V1beta1TFServingSpec
from kserve.models.v1beta1_torch_serve_spec import V1beta1TorchServeSpec
from kserve.models.v1beta1_transformer_config import V1beta1TransformerConfig
from kserve.models.v1beta1_transformer_spec import V1beta1TransformerSpec
from kserve.models.v1beta1_transformers_config import V1beta1TransformersConfig
from kserve.models.v1beta1_triton_spec import V1beta1TritonSpec
from kserve.models.v1beta1_xg_boost_spec import V1beta1XGBoostSpec



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import psutil
from typing import Dict, Union
from cloudevents.http import CloudEvent, to_binary, to_structured


def is_running_in_k8s():
    return os.path.isdir('/var/run/secrets/kubernetes.io/')


def get_current_k8s_namespace():
    with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace', 'r') as f:
        return f.readline()


def get_default_target_namespace():
    if not is_running_in_k8s():
        return 'default'
    return get_current_k8s_namespace()


def set_isvc_namespace(inferenceservice):
    isvc_namespace = inferenceservice.metadata.namespace
    namespace = isvc_namespace or get_default_target_namespace()
    return namespace


def cpu_count():
    """Get the available CPU count for this system.
    Takes the minimum value from the following locations:
    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    count = os.cpu_count()

    # Check CPU affinity if available
    try:
        affinity_count = len(psutil.Process().cpu_affinity())
        if affinity_count > 0:
            count = min(count, affinity_count)
    except Exception:
        pass

    # Check cgroups if available
    if sys.platform == "linux":
        try:
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us") as f:
                period = int(f.read())
            cgroups_count = int(quota / period)
            if cgroups_count > 0:
                count = min(count, cgroups_count)
        except Exception:
            pass

    return count


def is_structured_cloudevent(body: Dict) -> bool:
    """Returns True if the JSON request body resembles a structured CloudEvent"""
    return "time" in body \
           and "type" in body \
           and "source" in body \
           and "id" in body \
           and "specversion" in body \
           and "data" in body


def create_response_cloudevent(model_name: str, body: Union[Dict, CloudEvent], response: Dict,
                               binary_event=False) -> tuple:
    ce_attributes = {}

    if os.getenv("CE_MERGE", "false").lower() == "true":
        if binary_event:
            ce_attributes = body._attributes
            if "datacontenttype" in ce_attributes:  # Optional field so must check
                del ce_attributes["datacontenttype"]
        else:
            ce_attributes = body
            del ce_attributes["data"]

        # Remove these fields so we generate new ones
        del ce_attributes["id"]
        del ce_attributes["time"]

    ce_attributes["type"] = os.getenv("CE_TYPE", "io.kserve.inference.response")
    ce_attributes["source"] = os.getenv("CE_SOURCE", f"io.kserve.kfserver.{model_name}")

    event = CloudEvent(ce_attributes, response)

    if binary_event:
        event_headers, event_body = to_binary(event)
    else:
        event_headers, event_body = to_structured(event)

    return event_headers, event_body



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Union
import sys
import inspect
import json
import tornado.web
from tornado.httpclient import AsyncHTTPClient
from cloudevents.http import CloudEvent
from http import HTTPStatus
from enum import Enum
from kserve.utils.utils import is_structured_cloudevent
import grpc
from tritonclient.grpc import InferResult, service_pb2_grpc
from tritonclient.grpc.service_pb2 import ModelInferRequest, ModelInferResponse


PREDICTOR_URL_FORMAT = "http://{0}/v1/models/{1}:predict"
EXPLAINER_URL_FORMAT = "http://{0}/v1/models/{1}:explain"
PREDICTOR_V2_URL_FORMAT = "http://{0}/v2/models/{1}/infer"
EXPLAINER_V2_URL_FORMAT = "http://{0}/v2/models/{1}/explain"


class ModelType(Enum):
    EXPLAINER = 1
    PREDICTOR = 2


class PredictorProtocol(Enum):
    REST_V1 = "v1"
    REST_V2 = "v2"
    GRPC_V2 = "grpc-v2"


class ModelMissingError(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return self.path


class InferenceError(RuntimeError):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


# Model is intended to be subclassed by various components within KServe.
class Model:
    def __init__(self, name: str):
        self.name = name
        self.ready = False
        self.protocol = PredictorProtocol.REST_V1.value
        self.predictor_host = None
        self.explainer_host = None
        # The timeout matches what is set in generated Istio resources.
        # We generally don't want things to time out at the request level here,
        # timeouts should be handled elsewhere in the system.
        self.timeout = 600
        self._http_client_instance = None
        self._grpc_client_stub = None

    async def __call__(self, body, model_type: ModelType = ModelType.PREDICTOR):
        request = await self.preprocess(body) if inspect.iscoroutinefunction(self.preprocess) \
            else self.preprocess(body)
        request = self.validate(request)
        if model_type == ModelType.EXPLAINER:
            response = (await self.explain(request)) if inspect.iscoroutinefunction(self.explain) \
                else self.explain(request)
        elif model_type == ModelType.PREDICTOR:
            response = (await self.predict(request)) if inspect.iscoroutinefunction(self.predict) \
                else self.predict(request)
        else:
            raise NotImplementedError
        response = self.postprocess(response)
        return response

    @property
    def _http_client(self):
        if self._http_client_instance is None:
            self._http_client_instance = AsyncHTTPClient(max_clients=sys.maxsize)
        return self._http_client_instance

    @property
    def _grpc_client(self):
        if self._grpc_client_stub is None:
            # requires appending ":80" to the predictor host for gRPC to work
            if ":" not in self.predictor_host:
                self.predictor_host = self.predictor_host + ":80"
            _channel = grpc.aio.insecure_channel(self.predictor_host)
            self._grpc_client_stub = service_pb2_grpc.GRPCInferenceServiceStub(_channel)
        return self._grpc_client_stub

    def validate(self, request):
        if self.protocol == PredictorProtocol.REST_V2:
            if "inputs" in request and not isinstance(request["inputs"], list):
                raise tornado.web.HTTPError(
                    status_code=HTTPStatus.BAD_REQUEST,
                    reason="Expected \"inputs\" to be a list"
                )
        elif isinstance(request, Dict) or self.protocol == PredictorProtocol.REST_V1:
            if "instances" in request and not isinstance(request["instances"], list):
                raise tornado.web.HTTPError(
                    status_code=HTTPStatus.BAD_REQUEST,
                    reason="Expected \"instances\" to be a list"
                )
        return request

    def load(self) -> bool:
        """
        Load handler can be overridden to load the model from storage
        self.ready flag is used for model health check
        :return: bool
        """
        self.ready = True
        return self.ready

    async def preprocess(self, request: Union[Dict, CloudEvent]) -> Union[Dict, ModelInferRequest]:
        """
        The preprocess handler can be overridden for data or feature transformation.
        The default implementation decodes to Dict if it is a binary CloudEvent
        or gets the data field from a structured CloudEvent.
        :param request: Dict|CloudEvent|ModelInferRequest
        :return: Transformed Dict|ModelInferRequest which passes to predict handler
        """
        response = request

        if isinstance(request, CloudEvent):
            response = request.data
            # Try to decode and parse JSON UTF-8 if possible, otherwise
            # just pass the CloudEvent data on to the predict function.
            # This is for the cases that CloudEvent encoding is protobuf, avro etc.
            try:
                response = json.loads(response.decode('UTF-8'))
            except (json.decoder.JSONDecodeError, UnicodeDecodeError) as e:
                # If decoding or parsing failed, check if it was supposed to be JSON UTF-8
                if "content-type" in request._attributes and \
                        (request._attributes["content-type"] == "application/cloudevents+json" or
                         request._attributes["content-type"] == "application/json"):
                    raise tornado.web.HTTPError(
                        status_code=HTTPStatus.BAD_REQUEST,
                        reason=f"Failed to decode or parse binary json cloudevent: {e}"
                    )

        elif isinstance(request, dict):
            if is_structured_cloudevent(request):
                response = request["data"]

        return response

    def postprocess(self, response: Union[Dict, ModelInferResponse]) -> Dict:
        """
        The postprocess handler can be overridden for inference response transformation
        :param response: Dict|ModelInferResponse passed from predict handler
        :return: Dict
        """
        if isinstance(response, ModelInferResponse):
            response = InferResult(response)
            return response.get_response(as_json=True)
        return response

    async def _http_predict(self, request: Dict) -> Dict:
        predict_url = PREDICTOR_URL_FORMAT.format(self.predictor_host, self.name)
        if self.protocol == PredictorProtocol.REST_V2.value:
            predict_url = PREDICTOR_V2_URL_FORMAT.format(self.predictor_host, self.name)
        json_header = {'Content-Type': 'application/json'}
        response = await self._http_client.fetch(
            predict_url,
            method='POST',
            request_timeout=self.timeout,
            headers=json_header,
            body=json.dumps(request)
        )
        if response.code != 200:
            raise tornado.web.HTTPError(
                status_code=response.code,
                reason=response.body)
        return json.loads(response.body)

    async def _grpc_predict(self, request: ModelInferRequest) -> ModelInferResponse:
        async_result = await self._grpc_client.ModelInfer(request=request, timeout=self.timeout)
        return async_result

    async def predict(self, request: Union[Dict, ModelInferRequest]) -> Union[Dict, ModelInferResponse]:
        """
        The predict handler can be overridden to implement the model inference.
        The default implementation makes a call to the predictor if predictor_host is specified
        :param request: Dict|ModelInferRequest passed from preprocess handler
        :return: Dict|ModelInferResponse
        """
        if not self.predictor_host:
            raise NotImplementedError
        if self.protocol == PredictorProtocol.GRPC_V2.value:
            return await self._grpc_predict(request)
        else:
            return await self._http_predict(request)

    async def explain(self, request: Dict) -> Dict:
        """
        The explain handler can be overridden to implement the model explanation.
        The default implementation makes an call to the explainer if explainer_host is specified
        :param request: Dict passed from preprocess handler
        :return: Dict
        """
        if self.explainer_host is None:
            raise NotImplementedError
        explain_url = EXPLAINER_URL_FORMAT.format(self.explainer_host, self.name)
        if self.protocol == PredictorProtocol.REST_V2.value:
            explain_url = EXPLAINER_V2_URL_FORMAT.format(self.explainer_host, self.name)
        response = await self._http_client.fetch(
            url=explain_url,
            method='POST',
            request_timeout=self.timeout,
            body=json.dumps(request)
        )
        if response.code != 200:
            raise tornado.web.HTTPError(
                status_code=response.code,
                reason=response.body)
        return json.loads(response.body)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tornado.web
from kserve.handlers.base import BaseHandler
from kserve.model_repository import ModelRepository


class LivenessHandler(BaseHandler):  # pylint:disable=too-few-public-methods
    def get(self):
        self.write({"status": "alive"})


class HealthHandler(BaseHandler):
    def initialize(self, models: ModelRepository):
        self.models = models  # pylint:disable=attribute-defined-outside-init

    def get(self, name: str):
        model = self.models.get_model(name)
        if model is None:
            raise tornado.web.HTTPError(
                status_code=404,
                reason="Model with name %s does not exist." % name
            )

        if self.models.is_model_ready(name):
            self.write({
                "name": name,
                "ready": True
            })
        else:
            self.set_status(503)
            self.write({
                "name": name,
                "ready": False
            })



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from http import HTTPStatus

import tornado.web
from ray.serve.api import RayServeHandle

from kserve.handlers.base import HTTPHandler
from kserve.model import ModelType


class ExplainHandler(HTTPHandler):
    async def post(self, name: str):
        try:
            body = json.loads(self.request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        # call model locally or remote model workers
        model = self.get_model(name)
        if not isinstance(model, RayServeHandle):
            response = await model(body, model_type=ModelType.EXPLAINER)
        else:
            model_handle = model
            response = await model_handle.remote(body, model_type=ModelType.EXPLAINER)
        self.write(response)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import inspect
import tornado.web
from kserve.handlers.base import BaseHandler
from kserve.model_repository import ModelRepository


class LoadHandler(BaseHandler):
    def initialize(self, models: ModelRepository):  # pylint:disable=attribute-defined-outside-init
        self.models = models

    async def post(self, name: str):
        try:
            if inspect.iscoroutinefunction(self.models.load):
                await self.models.load(name)
            else:
                self.models.load(name)
        except Exception:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            raise tornado.web.HTTPError(
                status_code=500,
                reason=f"Model with name {name} is not ready. "
                       f"Error type: {ex_type} error msg: {ex_value}"
            )

        if not self.models.is_model_ready(name):
            raise tornado.web.HTTPError(
                status_code=503,
                reason=f"Model with name {name} is not ready."
            )
        self.write({
            "name": name,
            "load": True
        })


class UnloadHandler(BaseHandler):
    def initialize(self, models: ModelRepository):  # pylint:disable=attribute-defined-outside-init
        self.models = models

    def post(self, name: str):
        try:
            self.models.unload(name)
        except KeyError:
            raise tornado.web.HTTPError(
                status_code=404,
                reason="Model with name %s does not exist." % name
            )
        self.write({
            "name": name,
            "unload": True
        })


class ListHandler(BaseHandler):
    def initialize(self, models: ModelRepository):
        self.models = models  # pylint:disable=attribute-defined-outside-init

    def get(self):
        self.write({"models": list(self.models.get_models().keys())})



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest.mock as mock
import pytest
import kserve
import shutil


def create_mock_item(path):
    mock_obj = mock.MagicMock()
    mock_obj.name = path
    mock_obj.readall.return_value = b"test"
    return mock_obj


def create_mock_blob(mock_storage, paths):
    mock_objs = [create_mock_item(path) for path in paths]
    mock_container = mock.MagicMock()
    mock_container.walk_blobs.return_value = mock_objs
    mock_container.list_blobs.return_value = mock_objs
    mock_container.download_blob.return_value = mock_objs[0]
    mock_svc = mock.MagicMock()
    mock_svc.get_container_client.return_value = mock_container
    mock_storage.return_value = mock_svc
    return mock_storage, mock_container


def get_call_args(call_args_list):
    arg_list = []
    for call in call_args_list:
        args, _ = call
        arg_list.append(args)
    return arg_list


# pylint: disable=protected-access
@pytest.fixture(scope='session', autouse=True)
def test_cleanup():
    yield None
    # Will be executed after the last test
    shutil.rmtree('some', ignore_errors=True)
    shutil.rmtree('dest_path', ignore_errors=True)


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.BlobServiceClient')
def test_blob(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://kfserving.blob.core.windows.net/triton/simple_string/'
    paths = ['simple_string/1/model.graphdef', 'simple_string/config.pbtxt']
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)

    # when
    kserve.Storage._download_blob(blob_path, "dest_path")

    # then
    arg_list = get_call_args(mock_container.download_blob.call_args_list)
    assert set(arg_list) == set([('simple_string/1/model.graphdef',),
                                 ('simple_string/config.pbtxt',)])

    mock_storage.assert_called_with('https://kfserving.blob.core.windows.net',
                                    credential=None)


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.Storage._get_azure_storage_token')
@mock.patch('kserve.storage.BlobServiceClient')
def test_secure_blob(mock_storage, mock_get_token, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://kfsecured.blob.core.windows.net/triton/simple_string/'
    mock_get_token.return_value = "some_token"

    # when
    with pytest.raises(RuntimeError):
        kserve.Storage._download_blob(blob_path, "dest_path")

    # then
    mock_get_token.assert_called()
    arg_list = []
    for call in mock_storage.call_args_list:
        _, kwargs = call
        arg_list.append(kwargs)
    assert arg_list == [{'credential': 'some_token'}]


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.BlobServiceClient')
def test_deep_blob(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://accountname.blob.core.windows.net/container/some/deep/blob/path'
    paths = ['f1', 'f2', 'd1/f11', 'd1/d2/f21', 'd1/d2/d3/f1231', 'd4/f41']
    fq_item_paths = ['some/deep/blob/path/' + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    try:
        kserve.Storage._download_blob(blob_path, "some/dest/path")
    except OSError:  # Permissions Error Handling
        pass

    # then
    actual_calls = get_call_args(mock_container.download_blob.call_args_list)
    assert set(actual_calls) == set(expected_calls)


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.BlobServiceClient')
def test_blob_file(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://accountname.blob.core.windows.net/container/somefile.text'
    paths = ['somefile']
    fq_item_paths = paths
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)
    kserve.Storage._download_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.download_blob.call_args_list)
    assert actual_calls == expected_calls


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.BlobServiceClient')
def test_blob_fq_file(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://accountname.blob.core.windows.net/container/folder/somefile.text'
    paths = ['somefile']
    fq_item_paths = ['folder/' + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    kserve.Storage._download_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.download_blob.call_args_list)
    assert actual_calls == expected_calls


@mock.patch('kserve.storage.os.makedirs')
@mock.patch('kserve.storage.BlobServiceClient')
def test_blob_no_prefix(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = 'https://accountname.blob.core.windows.net/container/'
    paths = ['somefile.text', 'somefolder/somefile.text']
    fq_item_paths = ['' + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    kserve.Storage._download_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.download_blob.call_args_list)
    assert set(actual_calls) == set(expected_calls)



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
from typing import List, Optional, Dict, Union
import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.log
import asyncio
from tornado import concurrent

from .utils import utils

import kserve.handlers as handlers
from kserve import Model
from kserve.model_repository import ModelRepository
from ray.serve.api import Deployment, RayServeHandle
from ray import serve

DEFAULT_HTTP_PORT = 8080
DEFAULT_GRPC_PORT = 8081
DEFAULT_MAX_BUFFER_SIZE = 104857600

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('--http_port', default=DEFAULT_HTTP_PORT, type=int,
                    help='The HTTP Port listened to by the model server.')
parser.add_argument('--grpc_port', default=DEFAULT_GRPC_PORT, type=int,
                    help='The GRPC Port listened to by the model server.')
parser.add_argument('--max_buffer_size', default=DEFAULT_MAX_BUFFER_SIZE, type=int,
                    help='The max buffer size for tornado.')
parser.add_argument('--workers', default=1, type=int,
                    help='The number of works to fork')
parser.add_argument('--max_asyncio_workers', default=None, type=int,
                    help='Max number of asyncio workers to spawn')

args, _ = parser.parse_known_args()

tornado.log.enable_pretty_logging()


class ModelServer:
    def __init__(self, http_port: int = args.http_port,
                 grpc_port: int = args.grpc_port,
                 max_buffer_size: int = args.max_buffer_size,
                 workers: int = args.workers,
                 max_asyncio_workers: int = args.max_asyncio_workers,
                 registered_models: ModelRepository = ModelRepository()):
        self.registered_models = registered_models
        self.http_port = http_port
        self.grpc_port = grpc_port
        self.max_buffer_size = max_buffer_size
        self.workers = workers
        self.max_asyncio_workers = max_asyncio_workers
        self._http_server: Optional[tornado.httpserver.HTTPServer] = None

    def create_application(self):
        return tornado.web.Application([
            # Server Liveness API returns 200 if server is alive.
            (r"/", handlers.LivenessHandler),
            (r"/v2/health/live", handlers.LivenessHandler),
            (r"/v1/models",
             handlers.ListHandler, dict(models=self.registered_models)),
            (r"/v2/models",
             handlers.ListHandler, dict(models=self.registered_models)),
            # Model Health API returns 200 if model is ready to serve.
            (r"/v1/models/([a-zA-Z0-9_-]+)",
             handlers.HealthHandler, dict(models=self.registered_models)),
            (r"/v2/models/([a-zA-Z0-9_-]+)/status",
             handlers.HealthHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):predict",
             handlers.PredictHandler, dict(models=self.registered_models)),
            (r"/v2/models/([a-zA-Z0-9_-]+)/infer",
             handlers.PredictHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):explain",
             handlers.ExplainHandler, dict(models=self.registered_models)),
            (r"/v2/models/([a-zA-Z0-9_-]+)/explain",
             handlers.ExplainHandler, dict(models=self.registered_models)),
            (r"/v2/repository/models/([a-zA-Z0-9_-]+)/load",
             handlers.LoadHandler, dict(models=self.registered_models)),
            (r"/v2/repository/models/([a-zA-Z0-9_-]+)/unload",
             handlers.UnloadHandler, dict(models=self.registered_models)),
        ], default_handler_class=handlers.NotFoundHandler)

    def start(self, models: Union[List[Model], Dict[str, Deployment]], nest_asyncio: bool = False):
        if isinstance(models, list):
            for model in models:
                if isinstance(model, Model):
                    self.register_model(model)
                else:
                    raise RuntimeError("Model type should be Model")
        elif isinstance(models, dict):
            if all([isinstance(v, Deployment) for v in models.values()]):
                serve.start(detached=True, http_options={"host": "0.0.0.0", "port": 9071})
                for key in models:
                    models[key].deploy()
                    handle = models[key].get_handle()
                    self.register_model_handle(key, handle)
            else:
                raise RuntimeError("Model type should be RayServe Deployment")
        else:
            raise RuntimeError("Unknown model collection types")

        if self.max_asyncio_workers is None:
            # formula as suggest in https://bugs.python.org/issue35279
            self.max_asyncio_workers = min(32, utils.cpu_count()+4)

        self._http_server = tornado.httpserver.HTTPServer(
            self.create_application(), max_buffer_size=self.max_buffer_size)

        logging.info("Listening on port %s", self.http_port)
        self._http_server.bind(self.http_port)
        logging.info("Will fork %d workers", self.workers)
        self._http_server.start(self.workers)

        logging.info(f"Setting max asyncio worker threads as {self.max_asyncio_workers}")
        asyncio.get_event_loop().set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self.max_asyncio_workers))

        # Need to start the IOLoop after workers have been started
        # https://github.com/tornadoweb/tornado/issues/2426
        # The nest_asyncio package needs to be installed by the downstream module
        if nest_asyncio:
            import nest_asyncio
            nest_asyncio.apply()

        tornado.ioloop.IOLoop.current().start()

    def register_model_handle(self, name: str, model_handle: RayServeHandle):
        self.registered_models.update_handle(name, model_handle)
        logging.info("Registering model handle: %s", name)

    def register_model(self, model: Model):
        if not model.name:
            raise Exception(
                "Failed to register model, model.name must be provided.")
        self.registered_models.update(model)
        logging.info("Registering model: %s", model.name)



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from kubernetes.client import V1ServiceAccountList, V1ServiceAccount, V1ObjectMeta

from kserve.api.creds_utils import check_sa_exists


@mock.patch('kubernetes.client.CoreV1Api.list_namespaced_service_account')
def test_check_sa_exists(mock_client):
    # Mock kubernetes client to return 2 accounts
    accounts = V1ServiceAccountList(
        items=[V1ServiceAccount(metadata=V1ObjectMeta(name=n)) for n in ['a', 'b']]
    )
    mock_client.return_value = accounts

    # then a, b should exist, c should not
    assert check_sa_exists('kubeflow', 'a') is True
    assert check_sa_exists('kubeflow', 'b') is True
    assert check_sa_exists('kubeflow', 'c') is False



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

from kubernetes import client

from kserve import V1beta1PredictorSpec
from kserve import V1beta1TFServingSpec
from kserve import V1beta1InferenceServiceSpec
from kserve import V1beta1InferenceService
from kserve import KServeClient

kserve_client = KServeClient(config_file='./kserve/test/kubeconfig')

mocked_unit_result = \
    '''
{
    "api_version": "serving.kserve.io/v1beta1",
    "kind": "InferenceService",
    "metadata": {
        "name": "flower-sample",
        "namespace": "kubeflow"
    },
    "spec": {
        "predictor": {
            "tensorflow": {
                "storage_uri": "gs://kfserving-samples/models/tensorflow/flowers"
            }
        }
    }
}
 '''


def generate_inferenceservice():
    tf_spec = V1beta1TFServingSpec(
        storage_uri='gs://kfserving-samples/models/tensorflow/flowers')
    predictor_spec = V1beta1PredictorSpec(tensorflow=tf_spec)

    isvc = V1beta1InferenceService(
        api_version='serving.kserve.io/v1beta1',
        kind='InferenceService',
        metadata=client.V1ObjectMeta(name='flower-sample'),
        spec=V1beta1InferenceServiceSpec(predictor=predictor_spec))
    return isvc


def test_inferenceservice_client_creat():
    '''Unit test for kserve create api'''
    with patch('kserve.api.kserve_client.KServeClient.create',
               return_value=mocked_unit_result):
        isvc = generate_inferenceservice()
        assert mocked_unit_result == kserve_client.create(
            isvc, namespace='kubeflow')


def test_inferenceservice_client_get():
    '''Unit test for kserve get api'''
    with patch('kserve.api.kserve_client.KServeClient.get',
               return_value=mocked_unit_result):
        assert mocked_unit_result == kserve_client.get(
            'flower-sample', namespace='kubeflow')


def test_inferenceservice_client_watch():
    '''Unit test for kserve get api'''
    with patch('kserve.api.kserve_client.KServeClient.get',
               return_value=mocked_unit_result):
        assert mocked_unit_result == kserve_client.get('flower-sample', namespace='kubeflow',
                                                       watch=True, timeout_seconds=120)


def test_inferenceservice_client_patch():
    '''Unit test for kserve patch api'''
    with patch('kserve.api.kserve_client.KServeClient.patch',
               return_value=mocked_unit_result):
        isvc = generate_inferenceservice()
        assert mocked_unit_result == kserve_client.patch(
            'flower-sample', isvc, namespace='kubeflow')


def test_inferenceservice_client_replace():
    '''Unit test for kserve replace api'''
    with patch('kserve.api.kserve_client.KServeClient.replace',
               return_value=mocked_unit_result):
        isvc = generate_inferenceservice()
        assert mocked_unit_result == kserve_client.replace(
            'flower-sample', isvc, namespace='kubeflow')


def test_inferenceservice_client_delete():
    '''Unit test for kserve delete api'''
    with patch('kserve.api.kserve_client.KServeClient.delete',
               return_value=mocked_unit_result):
        assert mocked_unit_result == kserve_client.delete(
            'flower-sample', namespace='kubeflow')



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from http import HTTPStatus

import tornado.web

import cloudevents.exceptions as ce
from cloudevents.http import CloudEvent, from_http
from cloudevents.sdk.converters.util import has_binary_headers

from ray.serve.api import RayServeHandle

from kserve.handlers.base import HTTPHandler
from kserve.utils.utils import is_structured_cloudevent, create_response_cloudevent


class PredictHandler(HTTPHandler):
    def get_binary_cloudevent(self) -> CloudEvent:
        try:
            # Use default unmarshaller if contenttype is set in header
            if "ce-contenttype" in self.request.headers:
                event = from_http(self.request.headers, self.request.body)
            else:
                event = from_http(self.request.headers, self.request.body, lambda x: x)

            return event
        except (ce.MissingRequiredFields, ce.InvalidRequiredFields, ce.InvalidStructuredJSON,
                ce.InvalidHeadersFormat, ce.DataMarshallerError, ce.DataUnmarshallerError) as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Cloud Event Exceptions: %s" % e
            )

    async def post(self, name: str):
        is_cloudevent = False
        is_binary_cloudevent = False

        if has_binary_headers(self.request.headers):
            is_cloudevent = True
            is_binary_cloudevent = True
            body = self.get_binary_cloudevent()
        else:
            try:
                body = json.loads(self.request.body)
            except json.decoder.JSONDecodeError as e:
                raise tornado.web.HTTPError(
                    status_code=HTTPStatus.BAD_REQUEST,
                    reason="Unrecognized request format: %s" % e
                )

            if is_structured_cloudevent(body):
                is_cloudevent = True

        # call model locally or remote model workers
        model = self.get_model(name)
        if not isinstance(model, RayServeHandle):
            response = await model(body)
        else:
            model_handle = model
            response = await model_handle.remote(body)

        # if we received a cloudevent, then also return a cloudevent
        if is_cloudevent:
            headers, response = create_response_cloudevent(name, body, response, is_binary_cloudevent)

            if is_binary_cloudevent:
                self.set_header("Content-Type", "application/json")
            else:
                self.set_header("Content-Type", "application/cloudevents+json")

            for k, v in headers.items():
                self.set_header(k, v)

        self.write(response)



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

# KServe K8S constants
KSERVE_GROUP = 'serving.kserve.io'
KSERVE_KIND = 'InferenceService'
KSERVE_PLURAL = 'inferenceservices'
KSERVE_KIND_TRAINEDMODEL = 'TrainedModel'
KSERVE_PLURAL_TRAINEDMODEL = 'trainedmodels'
KSERVE_V1BETA1_VERSION = 'v1beta1'
KSERVE_V1ALPHA1_VERSION = "v1alpha1"

KSERVE_V1BETA1 = KSERVE_GROUP + '/' + KSERVE_V1BETA1_VERSION
KSERVE_V1ALPHA1 = KSERVE_GROUP + '/' + KSERVE_V1ALPHA1_VERSION

KSERVE_LOGLEVEL = os.environ.get('KSERVE_LOGLEVEL', 'INFO').upper()

# INFERENCESERVICE credentials common constants
INFERENCESERVICE_CONFIG_MAP_NAME = 'inferenceservice-config'
INFERENCESERVICE_SYSTEM_NAMESPACE = 'kserve'
DEFAULT_SECRET_NAME = "kserve-secret-"
DEFAULT_SA_NAME = "kserve-service-credentials"

# S3 credentials constants
S3_ACCESS_KEY_ID_DEFAULT_NAME = "AWS_ACCESS_KEY_ID"
S3_SECRET_ACCESS_KEY_DEFAULT_NAME = "AWS_SECRET_ACCESS_KEY"
S3_DEFAULT_CREDS_FILE = '~/.aws/credentials'

# GCS credentials constants
GCS_CREDS_FILE_DEFAULT_NAME = 'gcloud-application-credentials.json'
GCS_DEFAULT_CREDS_FILE = '~/.config/gcloud/application_default_credentials.json'

# Azure credentials constants
AZ_DEFAULT_CREDS_FILE = '~/.azure/azure_credentials.json'



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service import V1beta1InferenceService  # noqa: E501
from kserve.rest import ApiException

class TestV1beta1InferenceService(unittest.TestCase):
    """V1beta1InferenceService unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceService
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # model = kserve.models.v1beta1_inference_service.V1beta1InferenceService()  # noqa: E501
        if include_optional :
            return V1beta1InferenceService(
                api_version = '0', 
                kind = '0', 
                metadata = None, 
                spec = kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                    explainer = kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                        active_deadline_seconds = 56, 
                        affinity = None, 
                        aix = kserve.models.v1beta1_aix_explainer_spec.V1beta1AIXExplainerSpec(
                            args = [
                                '0'
                                ], 
                            command = [
                                '0'
                                ], 
                            config = {
                                'key' : '0'
                                }, 
                            env = [
                                None
                                ], 
                            env_from = [
                                None
                                ], 
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            ports = [
                                None
                                ], 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            type = '0', 
                            volume_devices = [
                                None
                                ], 
                            volume_mounts = [
                                None
                                ], 
                            working_dir = '0', ), 
                        alibi = kserve.models.v1beta1_alibi_explainer_spec.V1beta1AlibiExplainerSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            type = '0', 
                            working_dir = '0', ), 
                        automount_service_account_token = True, 
                        batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                            max_batch_size = 56, 
                            max_latency = 56, 
                            timeout = 56, ), 
                        canary_traffic_percent = 56, 
                        container_concurrency = 56, 
                        containers = [
                            None
                            ], 
                        dns_config = None, 
                        dns_policy = '0', 
                        enable_service_links = True, 
                        ephemeral_containers = [
                            None
                            ], 
                        host_aliases = [
                            None
                            ], 
                        host_ipc = True, 
                        host_network = True, 
                        host_pid = True, 
                        hostname = '0', 
                        image_pull_secrets = [
                            None
                            ], 
                        init_containers = [
                            None
                            ], 
                        logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                            mode = '0', 
                            url = '0', ), 
                        max_replicas = 56, 
                        min_replicas = 56, 
                        node_name = '0', 
                        node_selector = {
                            'key' : '0'
                            }, 
                        overhead = {
                            'key' : None
                            }, 
                        preemption_policy = '0', 
                        priority = 56, 
                        priority_class_name = '0', 
                        readiness_gates = [
                            None
                            ], 
                        restart_policy = '0', 
                        runtime_class_name = '0', 
                        scheduler_name = '0', 
                        security_context = None, 
                        service_account = '0', 
                        service_account_name = '0', 
                        share_process_namespace = True, 
                        subdomain = '0', 
                        termination_grace_period_seconds = 56, 
                        timeout = 56, 
                        tolerations = [
                            None
                            ], 
                        topology_spread_constraints = [
                            None
                            ], 
                        volumes = [
                            None
                            ], ), 
                    predictor = kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                        active_deadline_seconds = 56, 
                        affinity = None, 
                        automount_service_account_token = True, 
                        canary_traffic_percent = 56, 
                        container_concurrency = 56, 
                        dns_config = None, 
                        dns_policy = '0', 
                        enable_service_links = True, 
                        host_ipc = True, 
                        host_network = True, 
                        host_pid = True, 
                        hostname = '0', 
                        max_replicas = 56, 
                        min_replicas = 56, 
                        node_name = '0', 
                        onnx = kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), 
                        preemption_policy = '0', 
                        priority = 56, 
                        priority_class_name = '0', 
                        pytorch = kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0',
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), 
                        restart_policy = '0', 
                        runtime_class_name = '0', 
                        scheduler_name = '0', 
                        security_context = None, 
                        service_account = '0', 
                        service_account_name = '0', 
                        share_process_namespace = True, 
                        sklearn = kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), 
                        subdomain = '0', 
                        tensorflow = kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), 
                        termination_grace_period_seconds = 56, 
                        timeout = 56, 
                        triton = kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), 
                        xgboost = kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                            image = '0', 
                            image_pull_policy = '0', 
                            lifecycle = None, 
                            liveness_probe = None, 
                            name = '0', 
                            readiness_probe = None, 
                            resources = None, 
                            runtime_version = '0', 
                            security_context = None, 
                            startup_probe = None, 
                            stdin = True, 
                            stdin_once = True, 
                            storage_uri = '0', 
                            termination_message_path = '0', 
                            termination_message_policy = '0', 
                            tty = True, 
                            working_dir = '0', ), ), 
                    transformer = kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                        active_deadline_seconds = 56, 
                        affinity = None, 
                        automount_service_account_token = True, 
                        canary_traffic_percent = 56, 
                        container_concurrency = 56, 
                        dns_config = None, 
                        dns_policy = '0', 
                        enable_service_links = True, 
                        host_ipc = True, 
                        host_network = True, 
                        host_pid = True, 
                        hostname = '0', 
                        max_replicas = 56, 
                        min_replicas = 56, 
                        node_name = '0', 
                        preemption_policy = '0', 
                        priority = 56, 
                        priority_class_name = '0', 
                        restart_policy = '0', 
                        runtime_class_name = '0', 
                        scheduler_name = '0', 
                        security_context = None, 
                        service_account = '0', 
                        service_account_name = '0', 
                        share_process_namespace = True, 
                        subdomain = '0', 
                        termination_grace_period_seconds = 56, 
                        timeout = 56, ), ), 
                status = kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                    address = None, 
                    components = {
                        'key' : kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                            address = None, 
                            latest_created_revision = '0', 
                            latest_ready_revision = '0',
                            latest_rolledout_revision = '0',
                            previous_rolledout_revision = '0',
                            traffic = list(),
                            url = None, )
                        }, 
                    conditions = [
                        None
                        ], 
                    observed_generation = 56, 
                    url = None, )
            )
        else :
            return V1beta1InferenceService(
        )

    def testV1beta1InferenceService(self):
        """Test V1beta1InferenceService"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_paddle_server_spec import V1beta1PaddleServerSpec  # noqa: E501
from kserve.rest import ApiException

class TestV1beta1PaddleServerSpec(unittest.TestCase):
    """V1beta1PaddleServerSpec unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1PaddleServerSpec
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # model = kserve.models.v1beta1_paddle_server_spec.V1beta1PaddleServerSpec()  # noqa: E501
        if include_optional :
            return V1beta1PaddleServerSpec(
                args = [
                    '0'
                    ], 
                command = [
                    '0'
                    ], 
                env = [
                    None
                    ], 
                env_from = [
                    None
                    ], 
                image = '0', 
                image_pull_policy = '0', 
                lifecycle = None, 
                liveness_probe = None, 
                name = '0', 
                ports = [
                    None
                    ], 
                protocol_version = '0', 
                readiness_probe = None, 
                resources = None, 
                runtime_version = '0', 
                security_context = None, 
                startup_probe = None, 
                stdin = True, 
                stdin_once = True, 
                storage_uri = '0', 
                termination_message_path = '0', 
                termination_message_policy = '0', 
                tty = True, 
                volume_devices = [
                    None
                    ], 
                volume_mounts = [
                    None
                    ], 
                working_dir = '0'
            )
        else :
            return V1beta1PaddleServerSpec(
        )

    def testV1beta1PaddleServerSpec(self):
        """Test V1beta1PaddleServerSpec"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == '__main__':
    unittest.main()
